package de.ozgcloud.eingang.semantik.enginebased.dfoerdermittel;

import static org.assertj.core.api.Assertions.*;
import static org.assertj.core.api.InstanceOfAssertFactories.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.PostfachAddressTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;
import de.ozgcloud.eingang.semantik.enginebased.ServiceKontoBuildHelper;

class DFoerdermittelEngineBasedSemantikAdapterTest {

	@Spy
	@InjectMocks
	private DFoerdermittelEngineBasedSemantikAdapter adapter;

	@Mock
	private ServiceKontoBuildHelper serviceKontoHelper;

	@Nested
	class TestIsResponsible {

		@Test
		void shouldBeTrueWithFachnachricht() {
			var responsible = adapter.isResponsible(createWithFachnachricht());

			assertThat(responsible).isTrue();
		}

		@Test
		void shouldBeFalseForOuther() {
			var responsible = adapter.isResponsible(FormDataTestFactory.create());

			assertThat(responsible).isFalse();
		}
	}

	@Nested
	class TestParseFachnachricht {

		@Nested
		class ExtractData {
			@Test
			void shouldHaveFormData() {
				var result = adapter.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).isNotNull();
			}

			@Test
			void shouldHavePages() {
				var result = adapter.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).containsKey("Pages");
			}

			@Test
			void shouldHaveInboxRef() {
				var result = adapter.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).containsEntry("InboxReference", "sh/sh/4dd01647-b9d9-4775-1b50-08da3d83800a");
			}
		}

		@Nested
		class HandleFachnachrichtData {

			private Map<String, Object> extracted = Map.of("name", "Theo");

			@Test
			void shouldCallExtractData() {
				doReturn(extracted).when(adapter).extractFormDataFormXML(any());

				adapter.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				verify(adapter).extractFormDataFormXML(notNull());
			}

			@Test
			void shouldAddMap() {
				doReturn(extracted).when(adapter).extractFormDataFormXML(any());

				var result = adapter.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				assertThat(result.getFormData()).containsEntry("Fachnachricht", extracted);
			}

			@Test
			void shouldIgnoreEmptyData() {
				doReturn(Collections.emptyMap()).when(adapter).extractFormDataFormXML(any());

				var result = adapter.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				assertThat(result.getFormData()).doesNotContainKey("Fachnachricht");
			}
		}
	}

	FormData createWithFachnachricht() {

		return FormData.builder()
				.formData(Collections.emptyMap())
				.representation(createFachnachrichtFile()).build();
	}

	IncomingFile createFachnachrichtFile() {
		var tmpFile = TempFileUtils.writeTmpFile(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

		return IncomingFile.builder()
				.file(tmpFile)
				.contentType("application/xml")
				.size(1283)
				.name("Beispieldatensatz_Fachnachricht.xml")
				.build();
	}

	@Nested
	class TestProcessFachnachricht {

		@Mock
		FormData formData0;

		@Mock
		FormData formData1;

		@Mock
		FormData formData2;

		@Mock
		FormData formData3;

		@Mock
		FormData formData4;

		@BeforeEach
		void mock() {
			var formData = DFoerdermittelFormDataTestFactory.create();
			var fachnachrichtMap = DFoerdermittelFormDataTestFactory.createFachnachrichtMap();
			when(formData0.getFormData()).thenReturn(formData.getFormData());
			doReturn(formData1).when(adapter).addFormName(formData0);
			doReturn(formData2).when(adapter).addFormEngineName(formData1);
			doReturn(formData3).when(adapter).addServiceKonto(formData2, fachnachrichtMap);
			doReturn(formData4).when(adapter).addOrganisationsEinheitId(formData3, fachnachrichtMap);
		}

		@Test
		void shouldReturn() {
			var processedFormData = adapter.processFachnachricht(formData0);

			assertThat(processedFormData).isEqualTo(formData4);
		}
	}

	@Nested
	class TestAddServiceKonto {

		@BeforeEach
		void init() {
			when(serviceKontoHelper.buildOsiServiceKonto(any())).thenReturn(ServiceKontoTestFactory.create());
		}

		@Test
		void shouldHaveServiceKonto() {
			var formData = adapter.addServiceKonto(DFoerdermittelFormDataTestFactory.create(),
					DFoerdermittelFormDataTestFactory.createFachnachrichtMap());

			assertThat(formData.getHeader().getServiceKonto().getPostfachAddresses().get(0).getIdentifier())
					.asInstanceOf(type(StringBasedIdentifier.class)).extracting(StringBasedIdentifier::getPostfachId)
					.isEqualTo(PostfachAddressTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldRemovePrefix() {
			adapter.addServiceKonto(DFoerdermittelFormDataTestFactory.create(), DFoerdermittelFormDataTestFactory.createFachnachrichtMap());

			verify(serviceKontoHelper).buildOsiServiceKonto(DFoerdermittelFormDataTestFactory.POSTFACH_ID);
		}
	}

	@Nested
	class TestExtractPrefix {
		@Test
		void shouldRemoveAllBeforeLastSlash() {
			var result = adapter.extractPrefix("bla/bla/bla/12345");

			assertThat(result).isEqualTo("12345");
		}

		@Test
		void shouldBeFineWithoutSlash() {
			var result = adapter.extractPrefix("12345");

			assertThat(result).isEqualTo("12345");
		}
	}

	@Nested
	class TestAddOrganisationsEinheitId {
		@Test
		void shouldHaveOrganisationsEinheitId() {
			var formData = adapter.addOrganisationsEinheitId(DFoerdermittelFormDataTestFactory.create(),
					DFoerdermittelFormDataTestFactory.createFachnachrichtMap());

			assertThat(formData.getZustaendigeStelle()).isNotNull().extracting(ZustaendigeStelle::getOrganisationseinheitenId)
					.isEqualTo(DFoerdermittelFormDataTestFactory.ORGANISATIONS_EINHEIT_ID);
		}
	}

	@Nested
	class TestAddFormName {

		@Test
		void shouldHaveFormName() {
			var formData = adapter.addFormName(DFoerdermittelFormDataTestFactory.create());

			assertThat(formData.getHeader().getFormName()).isEqualTo("dFördermittelantrag");
		}

	}

	@Nested
	class TestAddFormEngineName {

		@Test
		void shouldHaveFormEngineName() {
			var formData = adapter.addFormEngineName(DFoerdermittelFormDataTestFactory.create());

			assertThat(formData.getHeader().getFormEngineName()).isEqualTo("dFördermittelantrag");
		}

	}
}
